#include <stdio.h>
#include <stdlib.h>

/* 
   Simple simulator for robot moving through an XMAX by YMAX grid.
   g[i][j] == 1 for walls, 0 for open space.
   rx,ry are the robots coordinates.
   There is a sensor in every non-wall square, that fires (value 1) 
   if it senses the robot, where the sensor error probabilities 
   are SENSOR_FALSE_POS and SENSOR_FALSE_NEG.
   After each step the robot decides whether or not to move with probability
   MOTION_PROB.  If it decides to move, then it moves with equality probability
   to any adjacent open square.
   This continues for TMAX steps.
   Data generated is a series of lines of the form:
        TIME_STEP  SENSOR_X_COORD  SENSOR_Y_COORD  SENSOR_READING
   The SENSOR_READING is 1 (fires) or 0 (does not fire) and is given for
   every sensor for every time slice.

   use: generate SEED MAPFILE DATAFILE TRUTHFILE FALSE_POS FALSE_NEG MOTION_PROB

   output: TRUTH is the file specifying robot's actual location
           DATA is the sensed data
*/


int XMAX;
int YMAX;
#define XINIT 0
#define YINIT 0
int TMAX;
int SENSOR_FALSE_POS;
int SENSOR_FALSE_NEG;
int MOTION_PROB;
int SEED;
char * TRUTH;
char * DATA;
char * MAP;

/*
int g[XMAX][YMAX] = {
    {0, 1, 0, 0, 0},
    {0, 0, 0, 1, 0},
    {1, 0, 1, 1, 0},
    {1, 0, 0, 0, 0},
    {1, 0, 0, 0, 1}};
*/
int g[100][100];

int rx = XINIT;
int ry = YINIT;

int main(int argc, char ** argv)
{
    int t, coin, sensed;
    int i,j;
    int nc;
    int choicex[4];
    int choicey[4];
    FILE * truth;
    FILE * data;
    FILE * map;

    if (argc != 9) { 
        printf("use: generate SEED MAP DATA TRUTH SENSOR_FALSE_POS SENSOR_FALSE_NEG MOTION_PROB TMAX\n");
        exit(1);
    }

    SEED = atoi(argv[1]);
    MAP = argv[2];
    DATA = argv[3];
    TRUTH = argv[4];
    SENSOR_FALSE_POS = atoi(argv[5]);
    SENSOR_FALSE_NEG = atoi(argv[6]);
    MOTION_PROB = atoi(argv[7]);
    TMAX = atoi(argv[8]);

    srandom(SEED);
    map = fopen(MAP, "r");
    truth = fopen(TRUTH, "w");
    data = fopen(DATA, "w");

    fscanf(map, "%i %i", &XMAX, &YMAX);
    for (i=0; i<XMAX; i++){
        for (j=0; j<YMAX; j++){
            fscanf(map, "%i", &(g[i][j]));
        }
    }

    fprintf(data, "%i %i\n", XMAX, YMAX);
    for (i=0; i<XMAX; i++){
        for (j=0; j<YMAX; j++){
            fprintf(data, " %1i", g[i][j]);
        }
        fprintf(data, "\n");
    }
    fprintf(data, "%f %f %f\n", SENSOR_FALSE_POS/100.0, SENSOR_FALSE_NEG/100.0, MOTION_PROB/100.0);
    fprintf(data, "%i\n", TMAX);

    for (t=1; t<=TMAX; t++){

        fprintf(truth, "%i %i %i\n", t, rx, ry);

        for (i=0; i<XMAX; i++){
            for (j=0; j<YMAX; j++){

                coin = (random()>>7)%100;

                if (g[i][j]==0){
                    if (rx==i && ry==j) {
                        if (coin < (100-SENSOR_FALSE_NEG))
                            sensed = 1;
                        else sensed = 0;
                    }
                    else {
                        if (coin < SENSOR_FALSE_POS)
                            sensed = 1;
                        else sensed = 0;
                    }
                    fprintf(data, "%d %d %d %d\n", t, i, j, sensed);
                }
            }
        }
        /* move robot */
        coin = (random()>>5)%100;
        if (coin < MOTION_PROB){

            nc = 0;
            if (rx>0 && g[rx-1][ry]==0) {choicex[nc]=rx-1; choicey[nc]=ry; nc++;}
            if (ry>0 && g[rx][ry-1]==0) {choicex[nc]=rx; choicey[nc]=ry-1; nc++;}
            if (rx<XMAX-1 && g[rx+1][ry]==0) {choicex[nc]=rx+1; choicey[nc]=ry; nc++;}
            if (ry<YMAX-1 && g[rx][ry+1]==0) {choicex[nc]=rx; choicey[nc]=ry+1; nc++;}
            coin = (random()>>9)%nc;
            rx = choicex[coin];
            ry = choicey[coin];

        }
    }
    return(0);
}
        


